//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// Universeller IR-Receiver UNIRV
//
// Kann fr nahezu jeden IR-Code/Fernbedienung verwendet werden.
// Wichtig: Zur Trgerfrequenz passenden Decoder-IC verwenden 
// (z.B. TSOP1136 - > 36 Khz ), auch die anderen Parameter sollten mglichst passen.
// Bei unpassendem Decoder-IC entstehen "Zufallscodes".
// 
//
// Implementation fr den STM32.
// Frank Bsing,  Januar 2010
//
// Used Periphals/Benutzte Peripherie:
// - Timer6 (kann beliebig gendert werden)
// - PB4    (kann beliebig gendert werden)
// - CRC
// - AFIO
//
// License:  GNU GENERAL PUBLIC LICENSE V3
//////////////////////////////////////////////////////////////////////////
//
//
// Bekannte Probleme:
// Gibt bei einigen Protokollen mehrere Codes aus 
// Lsung: Diese Ignorieren und nur den 1. Benutzen.
//
//////////////////////////////////////////////////////////////////////////
//
// Getestet mit: 
//    Sony, Onkyo, Panasonic, Sharp, Cyberhome, Samsung, Humax
 //   Cams: Nikon, Olympus  (DSLRs) (MINCODELEN verkleinern)
//
#include "stm32f4xx_conf.h"
#include "stm32f4xx_crc.h"
#include "ir.h"

//#include "led.h"
//////////////////////////////////////////////////////////////////////////

//Fr Test auskommentieren:
//volatile uint16_t times [IRTIMES];//(Debug)
//volatile short timescnt;//(Debug)

//////////////////////////////////////////////////////////////////////////
//Perpherieeinstellungen:

#define TIMER                 TIM2
#define TIMERPERIPH           RCC_APB1Periph_TIM2
#define TIMER_IRQHANDLER      TIM2_IRQHandler

#define GPIO                  GPIOE
#define GPIOPERIPH            RCC_AHB1Periph_GPIOE
#define GPIOPIN               GPIO_Pin_9
#define GPIO_PORTSOURCE       EXTI_PortSourceGPIOE
#define GPIO_PINSOURCE        EXTI_PinSource9

#define EXTI_LINE             EXTI_Line9
#define EXTI_IRQHANDLER       EXTI9_5_IRQHandler

#define NVIC_IRQCHANNELEXTI   EXTI9_5_IRQn
#define NVIC_IRQCHANNELTIM    TIM2_IRQn
  
//////////////////////////////////////////////////////////////////////////
//Empfngerkonstanten:

#define MINCODELEN          6
#define MINHALFBITLEN       200     //us
#define TIMEOUT             14000   

//////////////////////////////////////////////////////////////////////////
//Variable:

volatile uint32_t pressedkey;
volatile short timeovr; 
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// Gibt 32-Bit Code der gedrckten Taste zurck, 0 wenn keine Taste gedrckt wurde
// Returns 32Bit Code of pressed key or zero otherwise.

uint32_t ir_getkeypress(void)
{
    uint32_t tmp = pressedkey;
    pressedkey = 0;
    return tmp;
}

//////////////////////////////////////////////////////////////////////////
// Reset Hash

static inline void hash_reset(void) 
{
   CRC_ResetDR();
}
//////////////////////////////////////////////////////////////////////////
// Neuen Wert in den Hash

static inline void hash_calc(uint32_t dat) 
{
  CRC_CalcCRC(dat);
}
//////////////////////////////////////////////////////////////////////////
// Berechneter Hashwert

static inline uint32_t hash_get(void) 
{
  return CRC_GetCRC();
}

//////////////////////////////////////////////////////////////////////////
// "Working horse" Interrupt-handler: Aufgerufen bei Pinchange

void EXTI_IRQHANDLER(void)
{
  static uint16_t timestmp [IRTIMES];
  static uint32_t shortest = 0xffff;
  static short cnt;
  
  int time;
  int i;
  
  if(EXTI_GetITStatus(EXTI_LINE) != RESET) //On Pinchangeinterrupt :
  {

    time = TIMER->CNT;
    TIMER->CNT = 0;
            
    if (timeovr) {        //Timeout aufgetreten -> Empfangsende
        timeovr = 0;
        if (cnt > MINCODELEN) {
        
            shortest &= 0xfff0; //Klglicher Versuch, Jitter zu entfernen ;-)
        
            //Hash der normierten Zeiten berechnen:
            hash_reset();        
            for (i=2; i<cnt; i++) {
                int tmp;
                tmp = ((timestmp[i] &0xfff0) / shortest) & 0xfe; //Normierung mit Jitterentfernung. Hat noch Verbesserungpotential.
                hash_calc(tmp);  //Add to Hash
                //times[i] = tmp; //(Debug)
            }
            //timescnt = cnt;//(Debug)
        
            pressedkey = hash_get();
         }
     
         cnt = 0;
         shortest = 0xffff;
     
    } else if (time > MINHALFBITLEN) {
        timestmp[cnt] = time;        
        if (shortest > time) shortest = time; //Krzeste Zeit merken
        if (cnt < IRTIMES- 1) cnt++; 
    }
        
    //if (cnt & 1) LED_ON; else LED_OFF;//Toggle LED (Debug)
         
    EXTI_ClearITPendingBit(EXTI_LINE); //Reset interruptflag
  }
}

//////////////////////////////////////////////////////////////////////////
// Interrupt-handler: Timeroverflow
//
void TIMER_IRQHANDLER(void)
{

   if (TIM_GetITStatus(TIMER, TIM_IT_Update) == SET) {
   
        timeovr = 1;
        
        TIM_ClearITPendingBit(TIMER, TIM_IT_Update);
   }
}

//////////////////////////////////////////////////////////////////////////
// Peripherie-Init:
//
void init_ir (void)
{
  GPIO_InitTypeDef GPIO_InitStructure;  
  EXTI_InitTypeDef EXTI_InitStruct;
  NVIC_InitTypeDef NVIC_InitStructure;

  //ENABLE needed Periphals
  RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_CRC,ENABLE);
  RCC_APB1PeriphClockCmd(TIMERPERIPH,ENABLE);
  RCC_AHB1PeriphClockCmd(GPIOPERIPH,ENABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG,ENABLE);

  //GPIO-Config
  GPIO_InitStructure.GPIO_Pin = GPIOPIN; 
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_Init(GPIO, &GPIO_InitStructure);

  SYSCFG_EXTILineConfig(GPIO_PORTSOURCE, GPIO_PINSOURCE);

  //EXTI-Config:
  EXTI_InitStruct.EXTI_Line = EXTI_LINE;
  EXTI_InitStruct.EXTI_Mode = EXTI_Mode_Interrupt;
  EXTI_InitStruct.EXTI_Trigger = EXTI_Trigger_Rising_Falling;
  EXTI_InitStruct.EXTI_LineCmd = ENABLE;
  EXTI_Init(&EXTI_InitStruct);

  //NVIC-Config EXTI:
  NVIC_InitStructure.NVIC_IRQChannel = NVIC_IRQCHANNELEXTI;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0x00;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0x00;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
  
  //NVIC-Config Timer Overflow:
  NVIC_InitStructure.NVIC_IRQChannel = NVIC_IRQCHANNELTIM;
  NVIC_Init(&NVIC_InitStructure); 
  
  TIM_ITConfig(TIMER, TIM_IT_Update, ENABLE);
  TIMER->ARR = TIMEOUT;   
  TIMER->PSC = 72;      // Prescaler 
  TIMER->CNT = 0;  
  TIM_Cmd(TIMER, ENABLE);
}
