/*****************************************************
*
*	Control program for the PitSchuLight TV-Backlight
*	(c) Peter Schulten, Mlheim, Germany
*	peter_(at)_pitschu.de
*
*	Die unvernderte Wiedergabe und Verteilung dieses gesamten Sourcecodes
*	in beliebiger Form ist gestattet, sofern obiger Hinweis erhalten bleibt.
*
* 	Ich stelle diesen Sourcecode kostenlos zur Verfgung und biete daher weder
*	Support an noch garantiere ich fr seine Funktionsfhigkeit. Auerdem
*	bernehme ich keine Haftung fr die Folgen seiner Nutzung.

*	Der Sourcecode darf nur zu privaten Zwecken verwendet und modifiziert werden.
*	Darber hinaus gehende Verwendung bedarf meiner Zustimmung.
*
*	History
*	09.06.2013	pitschu		Start of work
*/

#include "hardware.h"
#include "ws2812.h"
#include "tvp5150_dcmi.h"
#include "stm32f4xx_flash.h"

/*
 * Funktionsweise:
 * Der TVP5150 liefert mit 27Mhz die Pixeldaten im Format ITU656. Hierbei weerden die Chromadaten CB und CR mit
 * den Lumadaten Y fr 2 Pixels in 4 bytes <Cb><y1><Cr><y2>. Die Daten werden per DCMI wie folgt verarbeitet:
 * ber die Crop Funktion des DCMI wird der gewnschte Pixelbereich eingestellt, die Crop Funktion des TVP5150
 * wird nicht genutzt, er liefert immer das volle Bild von 1728 bytes pro Zeile und 625 Zeilen pro Frame.
 *
 * Die Farb- und Helligkeitswerte der gelesenen Pixel werden zunchst in 48 x 28 Blcken aufaddiert (YCbCrSlots).
 * Da diese Verarbeitung fr eine Zeile lnger als 64us dauert, wird eine Zeile in 2 Hlften gelesen, die Verarbeitung
 * dauert dann nur 40us fr eine halbe Zeile.
 * Zunchst wird ber die Crop Einstellung die linke Hlfte des Bildes zeilenweise gelesen. Wenn das Bild bestehend
 * aus beiden PAL-Frames fertig ist (VSync), wird umgeschaltet auf die rechte Hlfte. Whrend der VSync Austastlcke
 * werden die YCbCr Summenblcke in RGB-Blcke umgerechnet und die YCbCr Daten wieder auf 0 initialisiert.
 *
 * Das DMA verwendet double buffering, so dass whrend der Aufaddition der YCbCr Daten aus Zeile N bereits die
 * Daten der Zeile N+1 per DMA in den anderen Buffer geladen werden.
 *
 * Die gesamte Verarbeitungskette luft wie folgt:
 *
 * 1. Daten werden per DMA geladen und nach HSync in <yCbCrSlots[48x28]> aufaddiert
 * 2. <yCbCr> Daten werden nach VSync in RGB <rgbSlots[48x28]> umgerechnet
 *
 * Bis hierher war alles whrend interupt Zeiten. Ab hier passiert es in Main()
 *
 * 3.  <rgbSlots[48x28] wird in eine variable Gre  <rgbPicture[Wid x High] umgerechnet. <Wid> x <Hih> kann
 *    von 1x1 bis 48x28 skaliert werden. Bei 1x1 bekommt man nur jeweils einen Farbwert fr jede Kante. 2x2 ist etwa so, wie
 *    das Original Ambilight.
 * 4. <rgbPicture> wird transformiert in RGB Werte fr den WS2812 Stripe <ws2812ledBuffer>.
 *    Pro Led gibts einen PID-Regler, allerdings mit zentral vorgegebenen pid Werten.
 * 5. Die RGB Werte der PIDs werden bitweise in Timer Werte umgerechnet und im DMA buffer fr den Timer abgelegt.
 *
 */
#define		PAL_WIDTH		360
#define		PAL_HEIGHT		288
#define		DMA_LINES		1

#undef 		HARD_SYNC
#define 	HARD_SYNC		// had no luck in using embedded codes; I use the hsync+vsync pins now


// 		some PAL timings: Front porch = 20 bytes, sync width = 128 bytes, back porch = 140 bytes
//		Vertical:	 front porch = 2.5 lines, sync width = 6 lines, back porch = 15 lines

unsigned short	captureWidth		= 712;				// in pixels
unsigned short	cropLeft			= 152;				// back porch bytes until visible area starts

unsigned short	cropTop				= 17;
unsigned short	cropHeight			= 286;

unsigned short	dmaWidth 	= 		PAL_WIDTH;			// in words (= DMA unit)
unsigned short	dmaLines	= 		DMA_LINES;			// currently we use only 1-line buffers
unsigned short	dmaBufLen	= 		PAL_WIDTH * 2 * DMA_LINES;

unsigned short	Brightness			= 123;
unsigned short	Color_saturation	= 128;
signed short	Hue_control			= 0;
unsigned short	Contrast			= 137;
unsigned short  col_g_r				= 0;
unsigned short  col_b_r				= 0;
unsigned short  col_r_g				= 0;
unsigned short  col_b_g				= 0;
unsigned short  col_r_b				= 0;
unsigned short  col_g_b				= 0;
unsigned short  wb_r				= 100;
unsigned short  wb_g				= 100;
unsigned short  wb_b				= 100;

unsigned short 	gamma_i				= 3;
unsigned short  factorI				= 6;

volatile short  for_det				= 0;
unsigned short  led_blank 			= 2;

volatile short 	captureReady 		= 0;		// semaphore for main(); set when frame captured completely

unsigned char	videoSourceSelect   = 0;		// 0 = auto; 1/2 = video channel
unsigned char	videoCurrentSource  = 1;		// current video channel set in TVP5150

typedef struct{
	unsigned short	captureWidth;
	unsigned short	cropLeft;
	unsigned short	cropTop;
	unsigned short	cropHeight;
	unsigned short	Brightness;
	unsigned short	Color_saturation;
	signed short	Hue_control;
	unsigned short	Contrast;
	unsigned short  col_g_r;
	unsigned short  col_b_r;
	unsigned short  col_r_g;
	unsigned short  col_b_g;
	unsigned short  col_r_b;
	unsigned short  col_g_b;
	unsigned short  wb_r;
	unsigned short  wb_g;
	unsigned short  wb_b;
	unsigned short 	led_blank;
	unsigned short  factorI;
} saveData_t;

volatile saveData_t saveData;

short 	TVP5150initRegisters(void);
void 	TVP5150initIOports(void);
void 	TVP5150initDCMI(void);
void 	TVP5150initDMA(void);

//------------------------------------------------------------------------------------------------------------

// typedef unsigned long YCbCr_t;
typedef struct {
	int8_t		Cb;
	uint8_t		Y0;
	int8_t		Cr;
	uint8_t		Y1;
} YCbCr_t;

static volatile YCbCr_t YCbCr_buf0 [(PAL_WIDTH * DMA_LINES)+1] = {{0}};		// the two alternating DMA buffers
static volatile YCbCr_t YCbCr_buf1 [(PAL_WIDTH * DMA_LINES)+1] = {{0}};


typedef struct {
	signed long		Y;
	signed long		Cb;
	signed long		Cr;
	signed long		cnt;			// number of pixels in this slot
} videoData_t;

static volatile videoData_t YCbCrSlots [LEDS_Y * LEDS_X] = {{0}};		// ycbcr values of frame currently captured
static volatile videoData_t *arrP;				// working pointer used in HSYNC handler

static unsigned short  arrXslot, arrYslot;			// number of pixels per array slot; set once when capture starts
static unsigned short  x_rest, y_rest, rest_y_cnt;
static unsigned short  b;				// counter within slice currently running

static int tvp5150_log_status(void);		// prints all TVP5150 registers

// rgbSlots[][] array is the main output of this module. It contains the raw RGB video data; its updated with each new frame
volatile rgbValue_t   rgbSlots [LEDS_Y][LEDS_X] = {{{0}}};					// RGB values are updated permanently

// black boarders of picture
unsigned short	dynLeft 	= 0;
unsigned short	dynRight 	= LEDS_X-1;
unsigned short	dynTop 		= 0;
unsigned short	dynBottom 	= LEDS_Y-1;


/***********************************************************************************/

short TVP5150init(void)
{
	TVP5150initIOports();
	TVP5150initDCMI();
	TVP5150initDMA();

	TVP5150initRegisters();

	delay_ms(1000);
	tvp5150_log_status();

	return(0);
}



void loadData_fun(void)
{
	uint16_t *index_s = det_index();
	uint16_t val, *save_p;
	long i;

	for(i=0;i<DAT_OFFSET;i++)
	{
		save_p = (uint16_t *)&saveData + i;
		val = *(index_s + i);
		*save_p = val;
	}

	captureWidth 	= saveData.captureWidth;
	cropLeft		= saveData.cropLeft;
	cropTop			= saveData.cropTop;
	cropHeight		= saveData.cropHeight;
	Brightness		= saveData.Brightness;
	Color_saturation= saveData.Color_saturation;
	Hue_control		= saveData.Hue_control;
	Contrast		= saveData.Contrast;
	col_g_r			= saveData.col_g_r;
	col_b_r			= saveData.col_b_r;
	col_r_g			= saveData.col_r_g;
	col_b_g			= saveData.col_b_g;
	col_r_b			= saveData.col_r_b;
	col_g_b			= saveData.col_g_b;
	wb_r			= saveData.wb_r;
	wb_g			= saveData.wb_g;
	wb_b			= saveData.wb_b;
	led_blank		= saveData.led_blank;
	factorI			= saveData.factorI;
}

void saveData_fun(uint16_t *index_s)
{
	long i;

	saveData.captureWidth		= captureWidth;
	saveData.cropLeft			= cropLeft;
	saveData.cropTop			= cropTop;
	saveData.cropHeight			= cropHeight;
	saveData.Brightness			= Brightness;
	saveData.Color_saturation	= Color_saturation;
	saveData.Hue_control		= Hue_control;
	saveData.Contrast			= Contrast;
	saveData.col_g_r			= col_g_r;
	saveData.col_b_r			= col_b_r;
	saveData.col_r_g			= col_r_g;
	saveData.col_b_g			= col_b_g;
	saveData.col_r_b			= col_r_b;
	saveData.col_g_b			= col_g_b;
	saveData.wb_r				= wb_r;
	saveData.wb_g				= wb_g;
	saveData.wb_b				= wb_b;
	saveData.led_blank			= led_blank;
	saveData.factorI			= factorI;

	uint16_t *pData = (uint16_t *)&saveData;

	FLASH_Unlock();
	for(i=0;i<DAT_OFFSET;i++)
	{
		FLASH_ProgramHalfWord((uint32_t)(index_s + i), pData[i]);
	}
	FLASH_Lock();
}

uint16_t* det_index(void)
{
	uint16_t *index_s = (uint16_t *)0x080E0000;
	uint16_t data = *index_s;

	if (data == 65535)
	{
		saveData_fun(index_s);
	}
	while (data!=65535)
	{
		data = *index_s;
		index_s = index_s + DAT_OFFSET;
	}
	return (index_s-2*DAT_OFFSET);
}


// Start the capturing. Captureing continues until TVP5150stopCapture() is called

void TVP5150startCapture(void)
{
	DCMI_CROPInitTypeDef DCMI_Crop;
	int timeout = 1000;

	captureReady = 0;
	arrP = &YCbCrSlots[0];

	DCMI_Crop.DCMI_CaptureCount 			= captureWidth*2 - 1;
	DCMI_Crop.DCMI_HorizontalOffsetCount 	= cropLeft;
	DCMI_Crop.DCMI_VerticalLineCount 		= cropHeight - 1;
	DCMI_Crop.DCMI_VerticalStartLine 		= cropTop;
	DCMI_CROPConfig (&DCMI_Crop);
	DCMI_CROPCmd (ENABLE);

	dmaWidth = captureWidth / 2;
	dmaBufLen = dmaLines * dmaWidth;

	arrXslot = (dmaWidth/2) / LEDS_X;
	x_rest = (dmaWidth/2) % LEDS_X;
	arrYslot = cropHeight / LEDS_Y;
	y_rest = cropHeight % LEDS_Y;
	b = arrYslot;
	rest_y_cnt = y_rest;

	DMA_SetCurrDataCounter(DMA2_Stream1, dmaBufLen);
	DMA_MemoryTargetConfig(DMA2_Stream1, (uint32_t)&YCbCr_buf0[0], DMA_Memory_0);
	DMA_DoubleBufferModeConfig (DMA2_Stream1, (uint32_t)&YCbCr_buf1[0], DMA_Memory_1);
	DMA_DoubleBufferModeCmd (DMA2_Stream1, ENABLE);

	DMA_Cmd(DMA2_Stream1, ENABLE);
	while (timeout && DMA_GetCmdStatus(DMA2_Stream1) != ENABLE)
		timeout--;

	DCMI_Cmd(ENABLE);
	DCMI_CaptureCmd(ENABLE);
}




void TVP5150stopCapture(void)
{
	int timeout = 1000;

	DCMI_CaptureCmd(DISABLE);
	DCMI_Cmd(DISABLE);
	DMA_Cmd(DMA2_Stream1, DISABLE);
	while (timeout && DMA_GetCmdStatus(DMA2_Stream1) != DISABLE)
		timeout--;
}




void TVP5150setPictureParams (void)
{
	I2C_WriteByte(TVP5150_I2C_ADR, R09_Brightness_control, 			(uint8_t)Brightness);
	I2C_WriteByte(TVP5150_I2C_ADR, R0A_Color_saturation_control, 	(uint8_t)Color_saturation);
	I2C_WriteByte(TVP5150_I2C_ADR, R0B_Hue_control, 				(uint8_t)Hue_control);
	I2C_WriteByte(TVP5150_I2C_ADR, R0C_Contrast_Control, 			(uint8_t)Contrast);
}



void TVP5150selectVideoSource (unsigned char src)
{
	if (src == 1)
		I2C_WriteByte(TVP5150_I2C_ADR, R00_Video_input_source_selection_1, 		0x00);		// select input #1
	else
		I2C_WriteByte(TVP5150_I2C_ADR, R00_Video_input_source_selection_1, 		0x02);		// select input #2

	videoCurrentSource = src;
}


// returns 0 when no video is detectod on active channel;
// returns 1 when video signal is detected and locked on active channel
unsigned char TVP5150hasVideoSignal ()
{
	int s = tvp5150_read(R88_Status_register_1);

	if ((s & 0x0e) != 0x0e)			// check color, Vsync and Hsync lock bits
		return (0);

	return 1;
}



unsigned char TVP5150getStatus1 ()
{
	int s = tvp5150_read(R88_Status_register_1);

	return (s & 0xff);
}



short TVP5150initRegisters(void)
{
	TVP5150selectVideoSource(videoCurrentSource);

	I2C_WriteByte(TVP5150_I2C_ADR, R01_Analog_channel_controls,					0b00010101);// AGC on
	I2C_WriteByte(TVP5150_I2C_ADR, R03_Miscellaneous_controls, 					0b10101111);// VBLK select and on, YCbCr mode, HSYNC,VSYNC,... enabled, Clock out enabled
	I2C_WriteByte(TVP5150_I2C_ADR, R0F_Configuration_shared_pins,				0b00000010);// VBLK instead of INTREQ

	TVP5150setPictureParams ();

//	I2C_WriteByte(TVP5150_I2C_ADR, R0D_Outputs_and_data_rates_select, 			0b00000000); // code range BT.601; 2s compl + offset;discrete sync output
	I2C_WriteByte(TVP5150_I2C_ADR, R0D_Outputs_and_data_rates_select, 			0b01100000); // extended code range (1..254); 2s compl;discrete sync output

	I2C_WriteByte(TVP5150_I2C_ADR, R11_Active_video_cropping_start_pixel_MSB,	0x00);
	I2C_WriteByte(TVP5150_I2C_ADR, R12_Active_video_cropping_start_pixel_LSB,	0x00);
	I2C_WriteByte(TVP5150_I2C_ADR, R13_Active_video_cropping_stop_pixel_MSB,	0x00);
	I2C_WriteByte(TVP5150_I2C_ADR, R14_Active_video_cropping_stop_pixel_LSB,	0x00);

	I2C_WriteByte(TVP5150_I2C_ADR, R16_Horizontal_sync_start,					0x80);

	I2C_WriteByte(TVP5150_I2C_ADR, R28_Video_standard,							0x00);		// Video standard autoswitch  mode
	return(0);
}




void TVP5150initDCMI(void)
{
	DCMI_CROPInitTypeDef DCMI_Crop;
	DCMI_CodesInitTypeDef DCMI_Codes;
	NVIC_InitTypeDef NVIC_InitStructure;
	DCMI_InitTypeDef DCMI_InitStruct;

	RCC_AHB2PeriphClockCmd(RCC_AHB2Periph_DCMI, ENABLE);

	DCMI_DeInit ();
	DCMI_InitStruct.DCMI_SynchroMode = 		DCMI_SynchroMode_Hardware;
	DCMI_InitStruct.DCMI_CaptureMode = 		DCMI_CaptureMode_SnapShot;	// will be reactivated in IRQ handler
	DCMI_InitStruct.DCMI_PCKPolarity = 		DCMI_PCKPolarity_Rising;
	DCMI_InitStruct.DCMI_VSPolarity = 		DCMI_VSPolarity_High;
	DCMI_InitStruct.DCMI_HSPolarity = 		DCMI_HSPolarity_High;
	DCMI_InitStruct.DCMI_CaptureRate = 		DCMI_CaptureRate_All_Frame;
	DCMI_InitStruct.DCMI_ExtendedDataMode = DCMI_ExtendedDataMode_8b;
	DCMI_Init(&DCMI_InitStruct);

	// see ITU BT.656 specs; currently not used in my implementation
	DCMI_Codes.DCMI_FrameStartCode = 		0xB6;
	DCMI_Codes.DCMI_LineStartCode = 		0x80;
	DCMI_Codes.DCMI_LineEndCode = 			0x9D;
	DCMI_Codes.DCMI_FrameEndCode = 			0xAB;
	DCMI_SetEmbeddedSynchroCodes (&DCMI_Codes);

	DCMI_Crop.DCMI_CaptureCount = 			captureWidth*2-1;
	DCMI_Crop.DCMI_HorizontalOffsetCount = 	cropLeft;
	DCMI_Crop.DCMI_VerticalLineCount = 		cropHeight-1;
	DCMI_Crop.DCMI_VerticalStartLine = 		cropTop;
	DCMI_CROPConfig (&DCMI_Crop);
	DCMI_CROPCmd (ENABLE);

	/* Enable DCMI Interrupt */
	NVIC_InitStructure.NVIC_IRQChannel = 	DCMI_IRQn;
	NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 1;
	NVIC_InitStructure.NVIC_IRQChannelSubPriority = 3;
	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&NVIC_InitStructure);

	DCMI_ClearFlag(DCMI_FLAG_VSYNCRI);
	DCMI_ClearFlag(DCMI_FLAG_LINERI);
	DCMI_ClearFlag(DCMI_FLAG_FRAMERI);
	DCMI_ClearFlag(DCMI_FLAG_ERRRI);
	DCMI_ClearFlag(DCMI_FLAG_OVFRI);

	DCMI_ITConfig (DCMI_IT_LINE, ENABLE);
	DCMI_ITConfig (DCMI_IT_VSYNC, ENABLE);
	DCMI_ITConfig (DCMI_IT_FRAME, ENABLE);
	DCMI_ITConfig (DCMI_IT_OVF, ENABLE);
}




void TVP5150initDMA(void)
{
	DMA_InitTypeDef  DMA_InitStructure;
	NVIC_InitTypeDef        NVIC_InitStructure;

	// Clock enable
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_DMA2, ENABLE);

	// DMA deinit
	DMA_Cmd(DMA2_Stream1, DISABLE);
	DMA_DeInit(DMA2_Stream1);

	// DMA init
	DMA_InitStructure.DMA_Channel 				= DMA_Channel_1;
	DMA_InitStructure.DMA_PeripheralBaseAddr 	= TVP5150_DCMI_REG_DR_ADDRESS;
	DMA_InitStructure.DMA_PeripheralInc 		= DMA_PeripheralInc_Disable;
	DMA_InitStructure.DMA_PeripheralDataSize 	= DMA_PeripheralDataSize_Word;
	DMA_InitStructure.DMA_PeripheralBurst 		= DMA_PeripheralBurst_Single;
	DMA_InitStructure.DMA_DIR 					= DMA_DIR_PeripheralToMemory;

	DMA_InitStructure.DMA_Memory0BaseAddr 		= (uint32_t)&YCbCr_buf0[0];
	DMA_InitStructure.DMA_BufferSize 			= dmaBufLen;
	DMA_InitStructure.DMA_MemoryInc 			= DMA_MemoryInc_Enable;
	DMA_InitStructure.DMA_MemoryDataSize 		= DMA_MemoryDataSize_Word;
	DMA_InitStructure.DMA_MemoryBurst 			= DMA_MemoryBurst_Single;

	DMA_InitStructure.DMA_Mode 					= DMA_Mode_Circular;
	DMA_InitStructure.DMA_Priority 				= DMA_Priority_High;
	DMA_InitStructure.DMA_FIFOMode 				= DMA_FIFOMode_Enable;
	DMA_InitStructure.DMA_FIFOThreshold 		= DMA_FIFOThreshold_Full;

	DMA_Init(DMA2_Stream1, &DMA_InitStructure);

	// we use double buffering. While first buffer is flled we can process second buffer
	DMA_DoubleBufferModeConfig (DMA2_Stream1, (uint32_t)&YCbCr_buf1[0], DMA_Memory_0);
	DMA_DoubleBufferModeCmd (DMA2_Stream1, ENABLE);

	/* Enable DMA interrupt: when line buffer is full, we are interupted */
	NVIC_InitStructure.NVIC_IRQChannel = DMA2_Stream1_IRQn;
	NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 1;
	NVIC_InitStructure.NVIC_IRQChannelSubPriority = 2;
	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&NVIC_InitStructure);

	DMA_ClearITPendingBit(DMA2_Stream1, DMA_IT_TCIF1);
	DMA_ITConfig (DMA2_Stream1, DMA_IT_TC, ENABLE);		// enable transfer complete event
}


//-------------------------------------------------------------------------------------------------------------------


// IRQ handler called when line buffer is full (just before HSYNC)
void DMA2_Stream1_IRQHandler (void)
{
	if (DMA_GetITStatus(DMA2_Stream1, DMA_IT_TCIF1) == SET)
	{
//		STM_EVAL_LEDOn(LED_RED);		// set check point for oszi
		DMA_ClearITPendingBit(DMA2_Stream1, DMA_IT_TCIF1);

		register videoData_t *p = (videoData_t*)arrP;
		register YCbCr_t	 *s = (YCbCr_t*)(DMA_GetCurrentMemoryTarget(DMA2_Stream1) == 0 ? &YCbCr_buf1[0] : &YCbCr_buf0[0]);
		register short a = arrXslot;
		short rest_x_cnt = x_rest;
		register short x;
		if (rest_x_cnt>0){
			a++; rest_x_cnt--;}

		for (x = dmaWidth; x != 0; x-=2)			// add video data in blocks (YCbCrSlots)
		{
			p->Cb += (signed long)s->Cb;			// Cb and Cr are 2s complement values
			p->Cr += (signed long)s->Cr;
			p->Y  += (signed long)s->Y0;			// add both Y values
			p->Y  += (signed long)s->Y1;
			p->cnt++;
			s+=2;
			if (--a == 0)
			{
				a = arrXslot;
				if (rest_x_cnt>0){
					a++; rest_x_cnt--;}
				p++;			// gather pixels into next X slot
			}
		}

		if (--b == 0)
		{
			b = arrYslot;
			if (rest_y_cnt>0){
				b++; rest_y_cnt--;}
			arrP += LEDS_X;			// points to next row
		}
	}
}




// IRQ handler called when VSYNC signal is active (field ends)
void DCMI_IRQHandler(void)
{
	// check masked IE bits
	if (DCMI->MISR & DCMI_IT_VSYNC)
	{
		DCMI_CROPInitTypeDef DCMI_Crop;
		short x,y,log;
		signed long l,g,r,b;

		// stop DMA, reset buffers to other half of picture and then restart
		DCMI_ClearFlag(DCMI_FLAG_VSYNCRI);
		DCMI_ClearITPendingBit(DCMI_IT_VSYNC);
		DCMI_CaptureCmd(DISABLE);
		DCMI_Cmd(DISABLE);

		DMA_Cmd(DMA2_Stream1, DISABLE);
		while (DMA_GetCmdStatus(DMA2_Stream1) != DISABLE);

		// reset values because they may have been changed by user
		dmaWidth = captureWidth / 2;
		dmaBufLen = dmaLines * dmaWidth;

		arrP = &YCbCrSlots[0];
		arrXslot = (dmaWidth/2) / LEDS_X;
		x_rest = (dmaWidth/2) % LEDS_X;
		arrYslot = cropHeight / LEDS_Y;
		y_rest = cropHeight % LEDS_Y;
		b = arrYslot;
		rest_y_cnt = y_rest;

		DCMI_CROPCmd (DISABLE);
		DCMI_Crop.DCMI_CaptureCount 			= captureWidth*2-1;
		DCMI_Crop.DCMI_HorizontalOffsetCount 	= cropLeft;
		DCMI_Crop.DCMI_VerticalLineCount 		= cropHeight - 1;
		DCMI_Crop.DCMI_VerticalStartLine 		= cropTop;
		DCMI_CROPConfig (&DCMI_Crop);
		DCMI_CROPCmd (ENABLE);

		DMA_SetCurrDataCounter(DMA2_Stream1, dmaBufLen);
		DMA_MemoryTargetConfig(DMA2_Stream1, (uint32_t)&YCbCr_buf0[0], DMA_Memory_0);
		DMA_DoubleBufferModeConfig (DMA2_Stream1, (uint32_t)&YCbCr_buf1[0], DMA_Memory_1);
		DMA_DoubleBufferModeCmd (DMA2_Stream1, ENABLE);
		DMA_ClearITPendingBit(DMA2_Stream1, DMA_IT_TCIF1);
		DMA_Cmd(DMA2_Stream1, ENABLE);
		while (DMA_GetCmdStatus(DMA2_Stream1) != ENABLE);

		DCMI_Cmd(ENABLE);
		DCMI_CaptureCmd(ENABLE);			// start capture

		// now transform the YCbCr values into RGB-values

		videoData_t *cp = (videoData_t *)&YCbCrSlots[0];

		for (y = 0; y < LEDS_Y; y++)
		{
			for (x = 0; x < LEDS_X; x++)
			{
				log = ((y<dynTop+3) || (y>dynBottom-3) || (x<dynLeft+3) || (x>dynRight-3) || (for_det)) && (cp->cnt > 0);
				if (log)
				{
					long Y = (cp->Y / cp->cnt);
					long Cb =(cp->Cb / cp->cnt);		// build average values
					long Cr =(cp->Cr / cp->cnt);

					Y = Y*500; 							// multiply Y, we have 2 Y values!

					g = Y - (Cr * 714 + Cb * 344);
					r = Cr * 1403 + Y;
					b = Cb * 1770 + Y;

					b = b/1000;
					if (b < 0) b = 0;
					if (b > 254) b = 254;
					rgbSlots[y][x].B = b;

					g = g/1000;
					if (g < 0) g = 0;
					if (g > 254) g = 254;
					rgbSlots[y][x].G = g;

					r = r/1000;
					if (r < 0) r = 0;
					if (r > 254) r = 254;
					rgbSlots[y][x].R = r;
				}
				cp->Cb = 0;
				cp->Cr = 0;
				cp->Y = 0;
				cp->cnt= 0;

				cp += 1;
			}
		}
		if (for_det)
		{
			// determine dynTop
			l = 1;			// is_black = yes
			dynTop = 0;
			for (y = 0; y < (LEDS_Y/2-5); y++)
			{
				for (x = 0; x < LEDS_X; x++)
				{
					if (rgbSlots[y][x].R>25 || rgbSlots[y][x].G>25 || rgbSlots[y][x].B>25)
					{
						l = 0; break;
					}
				}
				if (l==1)	{dynTop = dynTop+1;}
				else {break;}
			}
			// determine dynBottom
			l = 1;			// is_black = yes
			dynBottom = LEDS_Y-1;
			for (y = LEDS_Y-1; y > (LEDS_Y/2+4); y--)
			{
				for (x = 0; x < LEDS_X; x++)
				{
					if (rgbSlots[y][x].R>25 || rgbSlots[y][x].G>25 || rgbSlots[y][x].B>25)
					{
						l = 0; break;
					}
				}
				if (l==1)	{dynBottom = dynBottom-1;}
				else {break;}
			}
			// determine dynLeft
			l = 1;			// is_black = yes
			dynLeft = 0;
			for (x = 0; x < (LEDS_X/2-5); x++)
			{
				for (y = 0; y < LEDS_Y; y++)
				{
					if (rgbSlots[y][x].R>25 || rgbSlots[y][x].G>25 || rgbSlots[y][x].B>25)
					{
						l = 0; break;
					}
				}
				if (l==1)	{dynLeft = dynLeft+1;}
				else {break;}
			}
			// determine dynRight
			l = 1;			// is_black = yes
			dynRight = LEDS_X-1;
			for (x = LEDS_X-1; x > (LEDS_X/2+4); x--)
			{
				for (y = 0; y < LEDS_Y; y++)
				{
					if (rgbSlots[y][x].R>25 || rgbSlots[y][x].G>25 || rgbSlots[y][x].B>25)
					{
						l = 0; break;
					}
				}
				if (l==1)	{dynRight = dynRight-1;}
				else {break;}
			}
			for_det = 0;
//			dynRight = dynRight-1;
//			dynBottom = dynBottom-1;
//			dynLeft = dynLeft+1;
//			dynTop = dynTop+1;
		}
		captureReady = 1;			// semaphore for main loop to update LEDs
	}

	if (DCMI->MISR & DCMI_IT_LINE)
	{
		DCMI_ClearFlag(DCMI_FLAG_LINERI);
		DCMI_ClearITPendingBit(DCMI_IT_LINE);
	}

	if (DCMI->MISR & DCMI_IT_FRAME)
	{
		DCMI_ClearFlag(DCMI_FLAG_FRAMERI);
		DCMI_ClearITPendingBit(DCMI_IT_FRAME);
	}

	if (DCMI->MISR & DCMI_IT_OVF)
	{
		DCMI_ClearFlag(DCMI_FLAG_OVFRI);
		DCMI_ClearITPendingBit(DCMI_IT_OVF);
	}

	if (DCMI->MISR & DCMI_IT_ERR)
	{
		DCMI_ClearFlag(DCMI_FLAG_ERRRI);
		DCMI_ClearITPendingBit(DCMI_IT_ERR);
	}
}



// IRQ called when vertical blanking starts
// currently not used
void EXTI15_10_IRQHandler (void)
{
	if(EXTI_GetITStatus(EXTI_Line13) != RESET)		// V-Blank
	{
//		STM_EVAL_LEDToggle (LED_BLU);
		EXTI_ClearITPendingBit(EXTI_Line13);
	}
}




//-------------------------------------------------------------------------------------------------------------------

void TVP5150initIOports(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	EXTI_InitTypeDef   EXTI_InitStructure;
	NVIC_InitTypeDef   NVIC_InitStructure;

	//-----------------------------------------
	// Clock Enable Port-A, Port-B, Port-C und Port-E
	//-----------------------------------------
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOA,ENABLE);
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOB,ENABLE);
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOC,ENABLE);
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOE,ENABLE);

	// init D7 port as output and set it low. Pin is sampled during reset to get I2C adress bit 0
	GPIO_InitStructure.GPIO_Pin = 	GPIO_Pin_6;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_DOWN;
	GPIO_Init(GPIOE, &GPIO_InitStructure);

	/*
	  PE11 -> gpio (out)  = TVP5150 PDN
	  PE12 -> gpio (in)	  = TVP5150 AVID
	  PE13 -> gpio (in)	  = TVP5150 VBLK		( is IRQ source)
	  PE14 -> gpio (out)  = TVP5150 Reset
	  PE15 -> gpio (in)	  = TVP5150 FID
	 */

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_11 | GPIO_Pin_12;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_OD;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOE, &GPIO_InitStructure);

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_2;
	GPIO_Init(GPIOA, &GPIO_InitStructure);

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_14;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOE, &GPIO_InitStructure);

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_1;
	GPIO_Init(GPIOD, &GPIO_InitStructure);

	GPIO_ResetBits(GPIOE, GPIO_Pin_14);		// clear PDN power down
	GPIO_ResetBits(GPIOD, GPIO_Pin_1);		// activate RESET
	delay_ms (50);
	GPIO_SetBits(GPIOE, GPIO_Pin_14);			// release PDN
	delay_ms (50);
	GPIO_ResetBits(GPIOE, GPIO_Pin_6); 			// set D7 to 0
	delay_ms (50);
	GPIO_SetBits(GPIOD, GPIO_Pin_1);
	delay_ms (50);


	//-----------------------------------------
	// init port A
	//-----------------------------------------
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource4, GPIO_AF_DCMI); // PA4=DCMI_HSYNC -> HSNYC
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource6, GPIO_AF_DCMI); // PA6=DCMI_PCLK  -> PCLK

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_4 | GPIO_Pin_6;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_OD;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOA, &GPIO_InitStructure);

	//-----------------------------------------
	// init port B
	//-----------------------------------------
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource6, GPIO_AF_DCMI); // PB6=DCMI_D5    -> D5
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource7, GPIO_AF_DCMI); // PB7=DCMI_VSYNC -> VSYNC

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_6 | GPIO_Pin_7;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_OD;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOB, &GPIO_InitStructure);

	//-----------------------------------------
	// init port C
	//-----------------------------------------
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource6, GPIO_AF_DCMI); // PC6=DCMI_D0    -> D0
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource7, GPIO_AF_DCMI); // PC7=DCMI_D1    -> D1
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource8, GPIO_AF_DCMI); // PC8=DCMI_D2    -> D2
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource9, GPIO_AF_DCMI); // PC9=DCMI_D3    -> D3
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource11, GPIO_AF_DCMI); // PC11=DCMI_D4    -> D4

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_6 | GPIO_Pin_7 | GPIO_Pin_8 | GPIO_Pin_9 | GPIO_Pin_11;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_OD;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	//-----------------------------------------
	// init port E
	//-----------------------------------------
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource5, GPIO_AF_DCMI); // PE5=DCMI_D6    -> D6
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource6, GPIO_AF_DCMI); // PE6=DCMI_D7    -> D7

	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_5 | GPIO_Pin_6;

	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_OD;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOE, &GPIO_InitStructure);

	/* Connect EXTI Line13 to PE13 pin (V blanking) */
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE, EXTI_PinSource13);

	/* Configure EXTI Line13 */
	EXTI_InitStructure.EXTI_Line = EXTI_Line13;
	EXTI_InitStructure.EXTI_Mode = EXTI_Mode_Interrupt;
	EXTI_InitStructure.EXTI_Trigger = EXTI_Trigger_Rising;
	EXTI_InitStructure.EXTI_LineCmd = ENABLE;
	EXTI_Init(&EXTI_InitStructure);

	/* Enable and set EXTI Line13 Interrupt to the lowest priority */
	NVIC_InitStructure.NVIC_IRQChannel = EXTI15_10_IRQn;
	NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0x02;
	NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0x02;
	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&NVIC_InitStructure);
}


//---------------------------------------------------------------------------------------------------

int tvp5150_read (short addr)
{
	return (I2C_ReadByte(TVP5150_I2C_ADR, addr));
}



#ifdef USE_USB
static void dump_reg_range(char *s, u8 init, const u8 end, int max_line)
{
	int i = 0;

	while (init != (u8)(end + 1)) {
		if ((i % max_line) == 0) {
			if (i > 0)
				printf("\n");
			printf("tvp5150: %s reg 0x%02x = ", s, init);
		}
		printf("%02x ", tvp5150_read(init));

		init++;
		i++;
	}
	printf("\n");
}
#endif



static int tvp5150_log_status(void)
{
	#ifdef USE_USB
	printf("tvp5150: Video input source selection #1 = 0x%02x\n",
			tvp5150_read(R00_Video_input_source_selection_1));
	printf("tvp5150: Analog channel controls = 0x%02x\n",
			tvp5150_read(R01_Analog_channel_controls));
	printf("tvp5150: Operation mode controls = 0x%02x\n",
			tvp5150_read(R02_Operation_mode_controls));
	printf("tvp5150: Miscellaneous controls = 0x%02x\n",
			tvp5150_read(R03_Miscellaneous_controls));
	printf("tvp5150: Autoswitch mask= 0x%02x\n",
			tvp5150_read(R04_Autoswitch_mask));
	printf("tvp5150: Color killer threshold control = 0x%02x\n",
			tvp5150_read(R06_Color_killer_threshold_control));
	printf("tvp5150: Luminance processing controls #1 #2 and #3 = %02x %02x %02x\n",
			tvp5150_read(R07_Luminance_processing_control_1),
			tvp5150_read(R08_Luminance_processing_control_2),
			tvp5150_read(R0E_Luminance_processing_control_3));
	printf("tvp5150: Brightness control = 0x%02x\n",
			tvp5150_read(R09_Brightness_control));
	printf("tvp5150: Color saturation control = 0x%02x\n",
			tvp5150_read(R0A_Color_saturation_control));
	printf("tvp5150: Hue control = 0x%02x\n",
			tvp5150_read(R0B_Hue_control));
	printf("tvp5150: Contrast control = 0x%02x\n",
			tvp5150_read(R0C_Contrast_Control));
	printf("tvp5150: Outputs and data rates select = 0x%02x\n",
			tvp5150_read(R0D_Outputs_and_data_rates_select));
	printf("tvp5150: Configuration shared pins = 0x%02x\n",
			tvp5150_read(R0F_Configuration_shared_pins));
	printf("tvp5150: Active video cropping start = 0x%02x%02x\n",
			tvp5150_read(R11_Active_video_cropping_start_pixel_MSB),
			tvp5150_read(R12_Active_video_cropping_start_pixel_LSB));
	printf("tvp5150: Active video cropping stop  = 0x%02x%02x\n",
			tvp5150_read(R13_Active_video_cropping_stop_pixel_MSB),
			tvp5150_read(R14_Active_video_cropping_stop_pixel_LSB));
	printf("tvp5150: Genlock/RTC = 0x%02x\n",
			tvp5150_read(R15_Genlock_and_RTC));
	printf("tvp5150: Horizontal sync start = 0x%02x\n",
			tvp5150_read(R16_Horizontal_sync_start));
	printf("tvp5150: Vertical blanking start = 0x%02x\n",
			tvp5150_read(R18_Vertical_blanking_start));
	printf("tvp5150: Vertical blanking stop = 0x%02x\n",
			tvp5150_read(R19_Vertical_blanking_stop));
	printf("tvp5150: Chrominance processing control #1 and #2 = %02x %02x\n",
			tvp5150_read(R1A_Chrominance_control_1),
			tvp5150_read(R1B_Chrominance_control_2));
	printf("tvp5150: Interrupt reset register B = 0x%02x\n",
			tvp5150_read(R1C_Interrupt_reset_register_B));
	printf("tvp5150: Interrupt enable register B = 0x%02x\n",
			tvp5150_read(R1D_Interrupt_enable_register_B));
	printf("tvp5150: Interrupt configuration register B = 0x%02x\n",
			tvp5150_read(R1E_Interrupt_configuration_register_B));
	printf("tvp5150: Video standard = 0x%02x\n",
			tvp5150_read(R28_Video_standard));
	printf("tvp5150: Chroma gain factor: Cb=0x%02x Cr=0x%02x\n",
			tvp5150_read(R2C_Cb_gain_factor),
			tvp5150_read(R2D_Cr_gain_factor));
	printf("tvp5150: Macrovision on counter = 0x%02x\n",
			tvp5150_read(R2E_Macrovision_on_counter));
	printf("tvp5150: Macrovision off counter = 0x%02x\n",
			tvp5150_read(R2F_Macrovision_off_counter));
	printf("tvp5150: ITU-R BT.656.%d timing(TVP5150AM1 only)\n",
			(tvp5150_read(R30_656_revision_select) & 1) ? 3 : 4);
	printf("tvp5150: Device ID = %02x%02x\n",
			tvp5150_read(R80_Device_ID_MSB),
			tvp5150_read(R81_Device_ID_LSB));
	printf("tvp5150: ROM version = (hex) %02x.%02x\n",
			tvp5150_read(R82_ROM_version),
			tvp5150_read(R83_RAM_version_MSB));
	printf("tvp5150: Vertical line count = 0x%02x%02x\n",
			tvp5150_read(R84_Vertical_line_count_MSB),
			tvp5150_read(R85_Vertical_line_count_LSB));
	printf("tvp5150: Interrupt status register B = 0x%02x\n",
			tvp5150_read(R86_Interrupt_status_register_B));
	printf("tvp5150: Interrupt active register B = 0x%02x\n",
			tvp5150_read(R87_Interrupt_active_register_B));
	printf("tvp5150: Status regs #1 to #5 = %02x %02x %02x %02x %02x\n",
			tvp5150_read(R88_Status_register_1),
			tvp5150_read(R89_Status_register_2),
			tvp5150_read(R8A_Status_register_3),
			tvp5150_read(R8B_Status_register_4),
			tvp5150_read(R8C_Status_register_5));

	dump_reg_range("Teletext filter 1",   RB1_Teletext_filter_and_mask_1,
			RB1_Teletext_filter_and_mask_1 + 4, 8);
	dump_reg_range("Teletext filter 2",   RB6_Teletext_filter_and_mask_2,
			RB6_Teletext_filter_and_mask_2 + 4, 8);

	printf("tvp5150: Teletext filter enable = 0x%02x\n",
			tvp5150_read(RBB_Teletext_filter_control));
	printf("tvp5150: Interrupt status register A = 0x%02x\n",
			tvp5150_read(RC0_Interrupt_status_register_A));
	printf("tvp5150: Interrupt enable register A = 0x%02x\n",
			tvp5150_read(RC1_Interrupt_enable_register_A));
	printf("tvp5150: Interrupt configuration = 0x%02x\n",
			tvp5150_read(RC2_Interrupt_configuration_register_A));
	printf("tvp5150: VDP status register = 0x%02x\n",
			tvp5150_read(RC6_VDP_status));
	printf("tvp5150: FIFO word count = 0x%02x\n",
			tvp5150_read(RC7_FIFO_word_count));
	printf("tvp5150: FIFO interrupt threshold = 0x%02x\n",
			tvp5150_read(RC8_FIFO_interrupt_threshold));
	printf("tvp5150: FIFO reset = 0x%02x\n",
			tvp5150_read(RC9_FIFO_reset));
	printf("tvp5150: Line number interrupt = 0x%02x\n",
			tvp5150_read(RCA_Line_number_interrupt));
	printf("tvp5150: Pixel alignment register = 0x%02x%02x\n",
			tvp5150_read(RCC_Pixel_alignment_HSB),
			tvp5150_read(RCB_Pixel_alignment_LSB));
	printf("tvp5150: FIFO output control = 0x%02x\n",
			tvp5150_read(RCD_FIFO_output_control));
	printf("tvp5150: Full field enable = 0x%02x\n",
			tvp5150_read(RCF_Full_field_enable_));
	printf("tvp5150: Full field mode register = 0x%02x\n",
			tvp5150_read(RFC_Full_field_mode));

	dump_reg_range("CC   data",   R90_Closed_caption_data,
			R90_Closed_caption_data + 3, 8);

	dump_reg_range("WSS  data",   R94_WSS_CGMS_A_data,
			R94_WSS_CGMS_A_data + 5, 8);

	dump_reg_range("VPS  data",   R9A_VPS_Gemstar_2x_data,
			R9A_VPS_Gemstar_2x_data + 12, 8);

	dump_reg_range("VITC data",   RA7_VITC_data,
			RA7_VITC_data + 8, 10);

	dump_reg_range("Line mode",   RD0_Line_mode,
			RD0_Line_mode + 43, 8);
	#endif
	return 0;
}
